﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Windows.Threading;
using System.Threading;
using System.IO;

using HCW_Modify.Forms;
using HCW_Class;
using StopWatch;
using XImage_Class;

namespace HCW_Modify
{
	public partial class MDIMain:Form
	{

		#region HCWModify Program data definitions

		// Properties
		public bool text_flag = false;
		public bool gradient_flag = false;
		public HCW gradient_color = new HCW ();
		public int end_White = 254, end_Chroma = 192;
		public int current_x, current_y;
		public HCW current_color = new HCW ();
		public HCW sUpper_color = new HCW ();
		public HCW sLower_color = new HCW ();
		public HCW tUpper_color = new HCW ();
		public HCW tLower_color = new HCW ();
		public bool transform_flag = false;
		public bool transform_White = false;

		public string Open_Filter;	// defined for OpenFileDialog
		public string Save_Filter;	// defined for SaveFileDialog
		ImageCodecInfo[] Image_Encoders;

		public Blank_Form original_form;
		public Blank_Form modified_form;

		// GIF conversion palette
		int[] GIFG = { 0, 43, 85, 128, 170, 213, 255 };
		int[] GIFRB = { 0, 51, 102, 153, 204, 255 };

		private bool modified_flag = false;
		private Bitmap original_image;
		private string current_file_name = "";
		private string current_directory = "";

		// Current editing image in use
		public XImage current_image = new XImage (0, 0);
		
		// StopWatch global variables
		private Stop_Watch stopWatch = new Stop_Watch ();

		// Track pixels changed
		public long pixels_changed = 0;

		private long memory_last_used = 0;

		#endregion

		#region Main Form

		public MDIMain ()
		{
			InitializeComponent ();
		}

		private void MDIMain_Load (object sender, EventArgs e)
		{
			
			this.FormClosing += new System.Windows.Forms.FormClosingEventHandler (this.MainForm_FormClosing);

			GetFormats ();
		}

		void GetFormats ()
		{
			// Get the formats for file open and save
			Image_Encoders = ImageCodecInfo.GetImageEncoders ();
			foreach (ImageCodecInfo format in Image_Encoders)
			{
				Save_Filter += "|" + format.FormatDescription + "|" + format.FilenameExtension.ToLower ();
				Open_Filter += ";" + format.FilenameExtension.ToLower ();
			}
			Save_Filter = Save_Filter.Substring (1);
			Open_Filter = Open_Filter.Substring (1);
			Open_Filter = Save_Filter + "|Graphic Files (" + Open_Filter + ")|" + Open_Filter + "|All Files (*.*)|*.*";
		}

		private void MDIMain_Shown (object sender, EventArgs e)
		{

			//			MessageBox.Show (Environment.CurrentDirectory, "Environment.CurrentDirectory");

			String[] arguments = Environment.GetCommandLineArgs ();
			//			for (int i = 0;  i < arguments.Length; ++i)
			//				MessageBox.Show (arguments[i], "Environment.GetCommandLineArgs " + i);


			// If they drag and drop or give a command line argument then go directly to open
			if (1 < arguments.Length)
				FileOpen (arguments[1]);
		}

		private void MainForm_FormClosing (object sender, FormClosingEventArgs e)
		{
			DiscardChanges ();
		}

		#endregion

		#region File Menu Items

		private void HCW_PaletteMenuItem_Click (object sender, EventArgs e)
		{
			// Create a Hue/Chroma/White palette bitmap
			int x, y, i;
			Color pixel;
			HCW hcw = new HCW ();

			CloseForms ();

			// 1530 (6 * 255) max hue range, 511 max chroma then white range
			// 1530 hue, 1021 range up and down
			current_image.AllocateBitmap (1530, 1021);
			ReadyModified ();
			current_file_name = "HCW-Palette";
			ChangeImage (current_file_name);

			//			StopWatch (false);

			// Loop through the map pixels to create palette.
			for (x = 0; x < current_image.Width; x++)
			{
				hcw.Hue = (3600 * x) / 1530;
				for (y = 0; y < current_image.Height; y++)
				{
					i = y + 2 * (y / 510);
					switch (i / 256)
					{
						case 0:
							hcw.Chroma = y;
							break;
						case 1:
							hcw.White = y - 255;
							break;
						case 2:
							hcw.White = 765 - y;
							break;
						case 3:
						default:
							hcw.Chroma = 1020 - y;
							break;
					}
					pixel = hcw.RGB ();
					current_image.SetPixel (x, y, pixel);
				}
			}
			UpdateImage ();

			// StopWatch (true);

			ToolFileEnable ();
		}

		private void HW_PaletteMenuItem_Click (object sender, EventArgs e)
		{
			// Create a Hue/white range bitmap as a true hexagon
			int x = 0, y = 0, white;
			int i, j;
			Color pixel;
			HCW hcw = new HCW (Color.Red);

			CloseForms ();

			// 513 (256 + 255 + 2) max white x, 615 (256 + 255) * 6 / 5 + 2) max white y
			current_image.AllocateBitmap (513, 616);
			ReadyModified ();
			current_file_name = "HW-Palette";
			ChangeImage (current_file_name);

			// Loop through the white range pixels to create palette.
			for (white = 1; white < 256; white++)
			{
				int b = white * 6;
				hcw.White = (byte) (255 - white);
				for (i = 0; i < 6; ++i)
				{
					int t = i * white;
					for (j = 0; j < white; ++j)
					{
						hcw.Hue = (short) (1 + 3600 * (j + t) / b);
						pixel = hcw.RGB ();
						switch (i)
						{
							case 0:
								x = 256 + j;
								y = 307 - white * 6 / 5 + j * 6 / 10;
								break;
							case 1:
								x = 256 + white;
								y = 307 - white * 6 / 10 + j * 6 / 5;
								break;
							case 2:
								x = 256 + white - j;
								y = 307 + white * 6 / 10 + j * 6 / 10;
								break;
							case 3:
								x = 256 - j;
								y = 307 + white * 6 / 5 - j * 6 / 10;
								break;
							case 4:
								x = 256 - white;
								y = 307 + white * 6 / 10 - j * 6 / 5;
								break;
							case 5:
							default:
								x = 256 - white + j;
								y = 307 - white * 6 / 10 - j * 6 / 10;
								break;
						}
						current_image.SetPixel (x, y, pixel);
						current_image.SetPixel (x, y + 1, pixel);
					}
				}
			}
			UpdateImage ();
			ToolFileEnable ();
		}

		private void HC_PaletteMenuItem_Click (object sender, EventArgs e)
		{
			// Create a Hue/Chroma range bitmap
			int x = 0, y = 0, chroma;
			int i, j;
			Color pixel;
			HCW hcw = new HCW ();

			CloseForms ();

			// 513 (256 + 255 + 2) max chroma x, 615 (256 + 255) * 6 / 5 + 2) max chroma y
			current_image.AllocateBitmap (513, 616);
			ReadyModified ();
			current_file_name = "HC Palette";
			ChangeImage (current_file_name);

			// Loop through the chroma range pixels to create palette.
			for (chroma = 1; chroma < 256; chroma++)
			{
				int b = chroma * 6;
				hcw.Chroma = (byte) chroma;
				for (i = 0; i < 6; ++i)
				{
					int t = i * chroma;
					for (j = 0; j < chroma; ++j)
					{
						hcw.Hue = (short) (1 + 3600 * (j + t) / b);
						pixel = hcw.RGB ();
						switch (i)
						{
							case 0:
								x = 256 + j;
								y = 307 - chroma * 6 / 5 + j * 6 / 10;
								break;
							case 1:
								x = 256 + chroma;
								y = 307 - chroma * 6 / 10 + j * 6 / 5;
								break;
							case 2:
								x = 256 + chroma - j;
								y = 307 + chroma * 6 / 10 + j * 6 / 10;
								break;
							case 3:
								x = 256 - j;
								y = 307 + chroma * 6 / 5 - j * 6 / 10;
								break;
							case 4:
								x = 256 - chroma;
								y = 307 + chroma * 6 / 10 - j * 6 / 5;
								break;
							case 5:
							default:
								x = 256 - chroma + j;
								y = 307 - chroma * 6 / 10 - j * 6 / 10;
								break;
						}
						current_image.SetPixel (x, y, pixel);
						current_image.SetPixel (x, y + 1, pixel);
					}
				}
			}
			UpdateImage ();
			ToolFileEnable ();
		}

		private void roundSunburstMenuItem_Click (object sender, EventArgs e)
		{
			// Create a Custom Hue/white range bitmap as a true circle octants with bleed
			int x = 0, y = 0, white, decidegrees, rotate = 2400;
			double radians;
			Color pixel;
			HCW hcw = new HCW (Color.Red);

			CloseForms ();

			// 1001 (500 *2 + 1)
			current_image.AllocateBitmap (1001, 1001);
			ReadyModified ();
			current_file_name = "Round Sunburst";
			ChangeImage (current_file_name);

			// Loop through the white range pixels plus bleed.
			for (y = 1; y <= 500; ++y)
			{
				for (x = 0; x <= y; ++x)
				{
					radians = Math.Atan2 (x, y);
					decidegrees = (int) (1800 / Math.PI * radians);
					white = (int) (y / Math.Cos (radians));
					if (white > 255)
						white = 0;
					else
						white = 255 - white;

					// create circular rays
					if (3 > decidegrees % 50)
					{
						white = ((white + 63) / 64) * 64;
						if (0 == white)
							white = 16 + white;
					}

					hcw.White = white;
					hcw.Hue = rotate + decidegrees;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 + x, 500 - y, pixel);
					hcw.Hue = rotate + 900 - decidegrees;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 + y, 500 - x, pixel);
					hcw.Hue = rotate + decidegrees + 900;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 + y, 500 + x, pixel);
					hcw.Hue = rotate + 1800 - decidegrees;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 + x, 500 + y, pixel);
					hcw.Hue = rotate + decidegrees + 1800;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 - x, 500 + y, pixel);
					hcw.Hue = rotate + 2700 - decidegrees;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 - y, 500 + x, pixel);
					hcw.Hue = rotate + decidegrees + 2700;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 - y, 500 - x, pixel);
					hcw.Hue = rotate + 3600 - decidegrees;
					pixel = hcw.RGB ();
					current_image.SetPixel (500 - x, 500 - y, pixel);
				}
			}
			UpdateImage ();
			ToolFileEnable ();
		}

		private void hexSunburstMenuItem_Click (object sender, EventArgs e)
		{
			// Create a Custom Hue/white range bitmap as a flat approximate hexagon
			//
			// every color occupies two pixels in the x to create the hexagon
			// the edges in the corners are duplicated outward to fill in a square
			int x = 0, y = 0, white;
			int sextant, j, sex_white;
			int it;	// used to fill in the corners
			Color pixel;
			HCW hcw = new HCW (Color.Red);

			CloseForms ();

			// 1022 (512 + 510) max white x, 511 (256 + 255) max white y
			current_image.AllocateBitmap (1022, 511);
			ReadyModified ();
			current_file_name = "Hex Sunburst";
			ChangeImage (current_file_name);

			// Loop through the white range pixels to create palette.
			for (white = 1; white < 256; white++)
			{
				int white_6 = white * 6;
				int white_3 = white * 3;
				for (sextant = 0; sextant < 6; ++sextant)
				{
					sex_white = sextant * white;
					for (j = 0; j < white; ++j)
					{
						hcw.Hue = 600 + (3600 * (j + sex_white) + white_3) / white_6;
						hcw.White = 255 - white;

						int white_ray = 255 - white;
						if (0 == hcw.Hue % 30)
							white_ray = ((white_ray + 63) / 64) * 64;

						hcw.White = white_ray;

						pixel = hcw.RGB ();
						switch (sextant)
						{
							case 0: // left to up left
								x = 510 - white * 2 + j;
								y = 255 - j;


								// fill in corner
								if (255 == white)
								{
									for (it = 1; it <= x + 1 && it <= y; ++it)
									{
										// debug produce grayscale checkerboard
										//										int gray = (it % 4) * 64;
										//									Color pixel2 = Color.FromArgb (gray, gray, gray);

										// debug show image while drawing
										//										UpdateImage ();
										//										MessageBox.Show ("case 0");

										if (0 == x)
											current_image.SetPixel (x, y + it, pixel);
										if (it <= x)
											current_image.SetPixel (x - it, y - it, pixel);
										current_image.SetPixel (x - it + 1, y - it, pixel);
									}
								}


								break;
							case 1: // up left to up right
								x = 510 - white + j * 2;
								y = 255 - white;
								break;
							case 2: // up right
								x = 510 + white + j;
								y = 255 - white + j;


								// fill in corner
								if (255 == white)
								{
									for (it = 1; x + it < current_image.Width && it <= y; ++it)
									{
										current_image.SetPixel (x + it, y - it, pixel);
										if (x + it + 1 < current_image.Width)
											current_image.SetPixel (x + it + 1, y - it, pixel);
									}
								}


								break;
							case 3: // right
								x = 510 + white * 2 - j;
								y = 255 + j;

								// fill in corner
								if (255 == white)
								{
									for (it = 1; x + it < current_image.Width && y + it < current_image.Height; ++it)
									{
										if (current_image.Width == x + 2)
											current_image.SetPixel (x + it, y - it, pixel);
										current_image.SetPixel (x + it, y + it, pixel);
										if (x + it + 1 < current_image.Width)
											current_image.SetPixel (x + it + 1, y + it, pixel);
									}
								}


								break;
							case 4: // down right
								x = 510 + white - j * 2;
								y = 255 + white;
								break;
							case 5: // down left
							default:
								x = 510 - white - j;
								y = 255 + white - j;

								// fill in corner
								if (255 == white)
								{
									for (it = 1; it <= x + 1 && y + it < current_image.Height; ++it)
									{
										if (it <= x)
											current_image.SetPixel (x - it, y + it, pixel);
										current_image.SetPixel (x - it + 1, y + it, pixel);
									}
								}


								break;
						}
						current_image.SetPixel (x, y, pixel);
						current_image.SetPixel (x + 1, y, pixel);
					}
				}
			}
			UpdateImage ();
			ToolFileEnable ();
		}

		private void openMenuItem_Click (object sender, EventArgs e)
		{
			DiscardChanges ();
			CloseForms ();

			OpenFileDialog openFileDialog = new OpenFileDialog ();

			if (0 < current_directory.Length)
				openFileDialog.InitialDirectory = current_directory;
			else
				openFileDialog.InitialDirectory = Environment.GetFolderPath (Environment.SpecialFolder.Personal);

			openFileDialog.Filter = Open_Filter;
			openFileDialog.FilterIndex = Image_Encoders.Length + 1;
			if (openFileDialog.ShowDialog (this) != DialogResult.OK)
				return;

			string str = openFileDialog.FileName;
			openFileDialog = null;
			FileOpen (str);
		}

		private void FileOpen (string filename)
		{
			current_file_name = filename;
			current_directory = Path.GetFullPath (current_file_name);
			if (!File.Exists (current_file_name))
				return;
			Cursor.Current = Cursors.WaitCursor;
			original_image = new Bitmap (current_file_name);
			current_image.AllocateBitmap (original_image);

			// Status label of the filename
			toolStripStatusLabel.Text = current_file_name;

			// open original_form
			original_form = new Blank_Form ();
			original_form.MdiParent = this;
			original_form.ClientSize = new Size (current_image.Width, current_image.Height);
			original_form.BackgroundImage = original_image;
			original_form.Show ();

			// open modified_form
			ReadyModified ();

			RevertToOriginal ();

			OpenFileEnable (true);
		}

		private void ReadyModified ()
		{
			modified_form = new Blank_Form ();
			modified_form.MdiParent = this;
			modified_form.ClientSize = new Size (current_image.Width, current_image.Height);
			modified_form.BackgroundImage = current_image.Bitmap;
//			modified_form.BackgroundImage = original_image;
			modified_form.Show ();
		}

		private void SaveAsMenuItem_Click (object sender, EventArgs e)
		{
			FileSaveAs ();
		}

		private void FileSaveAs ()
		{
			string InitialDirectory, FileName, DefaultExt;
			int index;

			SaveFileDialog saveFileDialog = new SaveFileDialog ();
			saveFileDialog.CreatePrompt = false;
			saveFileDialog.OverwritePrompt = true;
			InitialDirectory = Path.GetDirectoryName (current_file_name) + "\\";
			saveFileDialog.InitialDirectory = InitialDirectory;
			DefaultExt = Path.GetExtension (current_file_name);
			if (0 == DefaultExt.Length)
				DefaultExt = ".PNG";
			DefaultExt = DefaultExt.Substring (1).ToLower ();
			saveFileDialog.DefaultExt = DefaultExt;
			for (index = 0; index < Image_Encoders.Length; ++index)
			{
				if (0 < Image_Encoders[index].FilenameExtension.IndexOf (DefaultExt))
					break;
			}
			saveFileDialog.FilterIndex = index + 1;
			FileName = Path.GetFileNameWithoutExtension (current_file_name) + "-2";

			saveFileDialog.FileName = FileName;
			saveFileDialog.Filter = Save_Filter;
			if (saveFileDialog.ShowDialog (this) != DialogResult.OK)
				return;

			index = saveFileDialog.FilterIndex - 1;

			Cursor.Current = Cursors.WaitCursor;
			FileName = saveFileDialog.FileName;

			modified_form.BackgroundImage.Save (FileName, Image_Encoders[index], null);
			modified_flag = false;
		}

		private void ExitToolsStripMenuItem_Click (object sender, EventArgs e)
		{
			DiscardChanges ();
			this.Close ();
		}

		#endregion

		#region Edit Menu Items

		private void blackenMenuItem_Click (object sender, EventArgs e)
		{
			int x, y;
			int noise_level;
			HCW hcw = new HCW ();

			noise_level = ReadInt ("Background (Black) noise removal level", 16, 32, 48);

			ChangeImage ("Blacken level " + noise_level);

			// Loop through the images pixels to black background
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));

					// Pure white is not allowed so a paint operation changing black to white
					// will not corrupt existing white values
					if (255 == hcw.White)
					{
						hcw.White = 254;
						//						hcw.RGB (Color.FromArgb (254, 254, 254));
						pixels_changed++;
					}
					else if (hcw.Chroma < noise_level)
					{
						hcw.Chroma = 0;
						pixels_changed++;
					}
					current_image.SetPixel (x, y, hcw.RGB ());
				}
			}
			UpdateImage ();
		}

		private void edgeMenuItem_Click (object sender, EventArgs e)
		{
			int x, y, pass;
			int edge_outline;
			HCW hcw = new HCW ();

			edge_outline = ReadInt ("Edge outline width", 1, 2, 9);

			ChangeImage ("Edge " + edge_outline);

			// Loop through the images pixels to outline edge.
			for (pass = edge_outline; pass > 0; --pass)
			{
				for (x = 1; x < current_image.Width - 1; x++)
				{
					for (y = 1; y < current_image.Height - 1; y++)
					{
						hcw.RGB (current_image.GetPixel (x, y));
						// is it a black spot?
						if (pass < hcw.Chroma)
						{
							Edge (current_image, x - 1, y - 1, pass);
							Edge (current_image, x - 1, y, pass);
							Edge (current_image, x - 1, y + 1, pass);
							Edge (current_image, x + 1, y - 1, pass);
							Edge (current_image, x + 1, y, pass);
							Edge (current_image, x + 1, y + 1, pass);
							Edge (current_image, x, y + 1, pass);
							Edge (current_image, x, y - 1, pass);
						}
					}
				}
			}
			UpdateImage ();
		}

		private void Edge (XImage image, int x, int y, int pass)
		{
			HCW hcw = new HCW (image.GetPixel (x, y));
			if (pass > hcw.Chroma)
			{
				image.SetPixel (x, y, Color.FromArgb (pass, pass, pass));
				pixels_changed++;
			}
		}

		private void toggleBWMenuItem_Click (object sender, EventArgs e)
		{
			int x, y;

			ChangeImage ("Toggle B/W");

			// Loop through the images pixels to toggle black and white
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{

					if (Color.White == current_image.GetPixel (x, y))
					{
						current_image.SetPixel (x, y, Color.Black);
						pixels_changed++;
					}
					else if (Color.Black == current_image.GetPixel (x, y))
					{
						current_image.SetPixel (x, y, Color.White);
						pixels_changed++;
					}
				}
			}
			UpdateImage ();
		}

		private void compressMenuItem_Click (object sender, EventArgs e)
		{
			int color_mask;

			color_mask = ReadInt ("Color mask", 2, 4, 64);
			//			RevertToOriginal ();
			//			StopWatch (false);
			//			Compress1 (color_mask);	// 243 msecs
			//			StopWatch (true);
			//			RevertToOriginal ();
			//			StopWatch (false);
			Compress2 (color_mask);	// 208 msecs
			//			StopWatch (true);
		}

		private void Compress1 (int color_mask)
		{
			int x, y;
			int r, g, b;
			Color pixel;

			ChangeImage ("Compress " + color_mask);

			// Loop through the images pixels
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{
					pixel = current_image.GetPixel (x, y);

					// compress color range values
					r = (pixel.R % color_mask);
					g = (pixel.G % color_mask);
					b = (pixel.B % color_mask);
					if (0 < (r + g + b))
					{
						r = pixel.R - r;
						g = pixel.G - g;
						b = pixel.B - b;
						pixel = Color.FromArgb (r, g, b);
						current_image.SetPixel (x, y, pixel);
						pixels_changed++;
					}
				}
			}
			UpdateImage ();
		}

		private void Compress2 (int color_mask)
		{
			int i;
			int r, g, b;

			ChangeImage ("Compress " + color_mask);

			int[] pixels = current_image.Array;

			// Loop through the images pixels
			for (i = 0; i < current_image.Length; i++)
			{
				b = pixels[i];
				g = (b >> 8) & 0xff;
				r = (b >> 16) & 0xff;
				b &= 0xff;

				// compress color range values
				if (255 > r)
				{
					r += color_mask / 2;
					r = r - r % color_mask;
				}
				if (255 > g)
				{
					g += color_mask / 2;
					g = g - g % color_mask;
				}
				if (255 > b)
				{
					b += color_mask / 2;
					b = b - b % color_mask;
				}
				b = b + (g << 8) + (r << 16);
				pixels[i] = (int) (0xff000000 + b);
			}
			UpdateImage ();
		}

		private void revertToOriginalMenuItem_Click (object sender, EventArgs e)
		{
			RevertToOriginal ();
		}

		private void RevertToOriginal ()
		{
			ChangeImage ("Unmodified Working Image");
			current_image.AllocateBitmap (original_image);
			UpdateImage ();
			modified_flag = false;
		}

		public void ChangeImage (string text)
		{
			// common function used to start change before image is edited
			Cursor.Current = Cursors.WaitCursor;
			modified_form.Text = text;
		}

		public void UpdateImage ()
		{
			// common function used after image update is edited
			modified_form.Invalidate ();
			modified_flag = true;
			if (0 < pixels_changed && displayPixelsChangedMenuItem.Checked)
				MessageBox.Show (String.Format ("{0:#,0}", pixels_changed), "Pixels changed");
			pixels_changed = 0;
		}

		private void closeMenuItem_Click (object sender, EventArgs e)
		{
			CloseForms ();
		}

		public void CloseForms ()
		{

			DiscardChanges ();
			if (null == modified_form)
				return;
			original_image = null;
//			current_image = null; reuse
			original_form = null;
			modified_form = null;
			foreach (Form childForm in MdiChildren)
			{
				childForm.Close ();
			}

			// Status label of the filename
			toolStripStatusLabel.Text = "Open a graphics file to modify.";
			OpenFileEnable (false);
		}

		private void gradientSmoothMenuItem_Click (object sender, EventArgs e)
		{
			int x, y, pass;
			int noise_level;
			int smoothing;

			noise_level = ReadInt ("Background (Black) noise level", 16, 32, 48);
			smoothing = ReadInt ("Gradient smoothing passes", 1, 10, 20);

			// Loop through the images pixels to smooth gradient.
			for (pass = smoothing; pass > 0; --pass)
			{
				ChangeImage ("Gradient smooth pass " + pass);
				for (x = 1; x < current_image.Width - 1; x++)
				{
					for (y = 1; y < current_image.Height - 1; y++)
					{
						// future anti-alias
						// ooo ooo ooo ooo ooo ooo ooo ooo oo. o.. oo. ooo ooo ooo
						// ooo ooo ooo oo. o.. ... ooo oo. o.. ... oo. ooo ooo ooo
						// oo. o.. ... ... ... ... oo. o.. ... ... ... ooo ooo ooo
						//              ^               ^
						Smooth (noise_level, current_image, x, y, 1, 0);
						Smooth (noise_level, current_image, x, y, 1, 1);
						Smooth (noise_level, current_image, x, y, 0, 1);
						Smooth (noise_level, current_image, x, y, -1, -1);
					}
				}
			}
			UpdateImage ();
		}

		private void Smooth (int noise_level, XImage image, int x, int y, int dx, int dy)
		{
			// smooth in RGB to avoid Hue wrapping
			Color rgb1, rgb2, rgb3;
			HCW hcw1, hcw2, hcw3;
			int r1, r2, r3;
			int g1, g2, g3;
			int b1, b2, b3;

			rgb1 = image.GetPixel (x - dx, y - dy);
			hcw1 = new HCW (rgb1);
			if (noise_level > hcw1.Chroma)
				return;
			rgb2 = image.GetPixel (x, y);
			hcw2 = new HCW (rgb2);
			if (noise_level > hcw2.Chroma)
				return;
			rgb3 = image.GetPixel (x + dx, y + dy);
			hcw3 = new HCW (rgb3);
			if (noise_level > hcw3.Chroma)
				return;

			r1 = rgb1.R;
			r2 = rgb2.R;
			r3 = rgb3.R;
			Slope (ref r1, ref r2, ref r3);

			g1 = rgb1.G;
			g2 = rgb2.G;
			g3 = rgb3.G;
			Slope (ref g1, ref g2, ref g3);

			b1 = rgb1.B;
			b2 = rgb2.B;
			b3 = rgb3.B;
			Slope (ref b1, ref b2, ref b3);

			image.SetPixel (x - dx, y - dy, Color.FromArgb (r1, g1, b1));
			image.SetPixel (x, y, Color.FromArgb (r2, g2, b2));
			image.SetPixel (x + dx, y + dy, Color.FromArgb (r3, g3, b3));
		}

		private void Slope (ref int v1, ref int v2, ref int v3)
		{
			// remove spikes because we expect a gradient.

			int a_slope1 = Math.Abs (v2 - v1);
			int s_slope1 = Math.Sign (v2 - v1);
			int a_slope2 = Math.Abs (v3 - v2);
			int s_slope2 = Math.Sign (v3 - v2);

			// insignificant slope
			if ((a_slope1 < 3) && (a_slope2 < 3))
				return;

			// spike?
			if ((s_slope1 != 0) && (s_slope1 != s_slope2))
			{
				Tilt (s_slope1, ref v1, ref v2, ref v3);
				return;
			}

			// insignificant change in slope as a percentage of slope
			//		if (Math.Abs (a_slope1 - a_slope2) < Math.Max (3, (a_slope1 + a_slope2) / 10))
			//			return;
			//
			//		if (a_slope1 < a_slope2)
			//		{
			//			Tilt (-s_slope1, ref v1, ref v2, ref v3);
			//		}
			//		else
			//		{
			//			Tilt (s_slope1, ref v1, ref v2, ref v3);
			//		}
			//		return;

		}

		private void Tilt (int slope, ref int v1, ref int v2, ref int v3)
		{

			v1 += slope;
			if ((0 > v1) || (255 < v1))
			{
				v2 += slope;
				v1 -= slope;
			}
			v2 -= 2 * slope;
			v3 += slope;
			if ((0 > v3) || (255 < v3))
			{
				v2 += slope;
				v3 -= slope;
			}
			pixels_changed++;
		}

		private void blackWhiteMenuItem_Click (object sender, EventArgs e)
		{
			int x, y;
			int white_level;
			HCW hcw = new HCW ();

			white_level = ReadInt ("White level", 64, 128, 254);

			ChangeImage ("Black/White level " + white_level);

			// Loop through the images pixels to convert to B/W
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));

					if (hcw.Chroma > 0)
						if (hcw.White < white_level)
						{
							current_image.SetPixel (x, y, Color.Black);
							pixels_changed++;
						}
						else if (hcw.White < 255)
						{
							current_image.SetPixel (x, y, Color.White);
							pixels_changed++;
						}
				}
			}
			UpdateImage ();
		}

		private void overlappingMenuItem_Click (object sender, EventArgs e)
		{
			int x, y, r;//, g, b;
			Color pixel;

			ChangeImage ("Grayscale (Overlapping)");

			// Loop through the images pixels to convert to Grayscale
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					pixel = current_image.GetPixel (x, y);
					//					r = pixel.R * 30 / 100;
					//					g = pixel.G * 59 / 100;
					//					b = pixel.B * 11 / 100;
					//					r += g + b;
					r = (int) (pixel.R * 30.0 / 100 + pixel.G * 59.0 / 100 + pixel.B * 11.0 / 100);
					current_image.SetPixel (x, y, Color.FromArgb (r, r, r));
				}
			UpdateImage ();
		}

		private void reversibleMenuItem_Click (object sender, EventArgs e)
		{
			int x, y, r, g, b, w;
			Color pixel;

			ChangeImage ("Grayscale (Reversible)");

			// Loop through the images pixels to convert to a reversible Grayscale
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					pixel = current_image.GetPixel (x, y);
					r = pixel.R / 32;
					g = pixel.G / 64;
					b = pixel.B / 32;
					w = 0;
					if (0 < (g & 1))
						w |= 1;
					if (0 < (b & 1))
						w |= 2;
					if (0 < (r & 1))
						w |= 4;
					if (0 < (g & 2))
						w |= 8;
					if (0 < (b & 2))
						w |= 16;
					if (0 < (r & 2))
						w |= 32;
					if (0 < (b & 4))
						w |= 64;
					if (0 < (r & 4))
						w |= 128;
					current_image.SetPixel (x, y, Color.FromArgb (w, w, w));
				}
			UpdateImage ();
		}

		private void undoReverseMenuItem_Click (object sender, EventArgs e)
		{
			int x, y, r, g, b, w;
			Color pixel;

			ChangeImage ("Grayscale (Undo)");

			// Loop through the images pixels to convert to a reversible Grayscale
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					pixel = current_image.GetPixel (x, y);
					r = pixel.R;
					g = pixel.G;
					b = pixel.B;
					if (r != g | r != b)
					{
						MessageBox.Show ("Not grayscale!");
						return;
					}
					w = r;
					r = 0;
					g = 0;
					b = 0;
					if (0 < (w & 1))
						g |= 1;
					if (0 < (w & 2))
						b |= 1;
					if (0 < (w & 4))
						r |= 1;
					if (0 < (w & 8))
						g |= 2;
					if (0 < (w & 16))
						b |= 2;
					if (0 < (w & 32))
						r |= 2;
					if (0 < (w & 64))
						b |= 4;
					if (0 < (w & 128))
						r |= 4;
					r *= 32;
					g *= 64;
					b *= 32;
					current_image.SetPixel (x, y, Color.FromArgb (r, g, b));
				}
			UpdateImage ();
		}

		private void gIFPaletteColorConversionMenuItem_Click (object sender, EventArgs e)
		{
			// Perform a GIF palette Color Conversion to the nearest color
			int x, y, r, g, b;
			Color pixel;

			ChangeImage ("GIF Conversion");

			// Loop through the images pixels to convert to nearest color
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					pixel = current_image.GetPixel (x, y);
					r = GIFRB[(pixel.R + 25) / 51];
					g = GIFG[(pixel.G * 10 + 210) / 425];
					b = GIFRB[(pixel.B + 25) / 51];
					current_image.SetPixel (x, y, Color.FromArgb (r, g, b));
				}
			UpdateImage ();
		}

		private void chromaMenuItem_Click (object sender, EventArgs e)
		{
			// Perform a Grayscale conversion by copying Chroma level to White
			// Primary colors all turn white
			int x, y;
			HCW hcw = new HCW ();

			ChangeImage ("Grayscale Chroma");
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));
					hcw.W = hcw.C;
					current_image.SetPixel (x, y, hcw.RGB ());
				}
			UpdateImage ();
		}

		private void whiteToChromaMenuItem_Click (object sender, EventArgs e)
		{
			// Perform a Grayscale conversion by copying White level to Chroma
			int x, y;
			HCW hcw = new HCW ();

			ChangeImage ("Grayscale White");
			for (x = 0; x < current_image.Width; x++)
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));
					hcw.C = hcw.W;
					current_image.SetPixel (x, y, hcw.RGB ());
				}
			UpdateImage ();
		}

		private void hueGradientMenuItem_Click (object sender, EventArgs e)
		{
			// Generate Gradient bitmap
			int x, y;
			double uH, uC, uW;
			double urH, urC, urW;
			double lH, lC, lW;
			double lrH, lrC, lrW;
			double tH, tC, tW;
			double trH, trC, trW;
			Color pixel;
			HCW hcw = new HCW (current_color);

			CloseForms ();

			Gradient gradient_form = new Gradient ();
			gradient_form.theParent = this;
			gradient_form.ShowDialog ();

			lH = sLower_color.Hue;
			lrH = (sUpper_color.Hue - sLower_color.Hue) / 255.0;
			lC = sLower_color.Chroma;
			lrC = (sUpper_color.Chroma - sLower_color.Chroma) / 255.0;
			lW = sLower_color.White;
			lrW = (sUpper_color.White - sLower_color.White) / 255.0;
			uH = tLower_color.Hue;
			urH = (tUpper_color.Hue - tLower_color.Hue) / 255.0;
			uC = tLower_color.Chroma;
			urC = (tUpper_color.Chroma - tLower_color.Chroma) / 255.0;
			uW = tLower_color.White;
			urW = (tUpper_color.White - tLower_color.White) / 255.0;

			current_image = new XImage (256, 256);
			ReadyModified ();
			current_file_name = "Hue Gradient";
			ChangeImage (current_file_name);

			// Loop through the range pixels to create gradient.
			for (x = 0; x < current_image.Width; x++)
			{
				tH = lH;
				trH = (uH - lH) / 256.0;
				tC = lC;
				trC = (uC - lC) / 256.0;
				tW = lW;
				trW = (uW - lW) / 256.0;
				for (y = 0; y < current_image.Width; ++y)
				{
					hcw.iHCW ((short) tH, (byte) tC, (byte) tW);
					pixel = hcw.RGB ();
					current_image.SetPixel (x, y, pixel);
					tH += trH;
					tC += trC;
					tW += trW;
				}
				uH += urH;
				uC += urC;
				uW += urW;
				lH += lrH;
				lC += lrC;
				lW += lrW;
			}
			UpdateImage ();
			ToolFileEnable ();
		}

		private void TransformMenuItem_Click (object sender, EventArgs e)
		{
			Transform transform_form = new Transform ();
			transform_form.ShowDialog (this);
		}

		private void fillGradientMenuItem_Click (object sender, EventArgs e)
		{
			// Fill Gradient on current bitmap current hue left shade to white right dark to edge
			gradient_flag = true;
			toolStripStatusLabel.Text = "Click the area for gradient fill.";
		}

		public void Gradient ()
		{
			// Fill Gradient on current bitmap current hue left shade to white right dark to edge
			int left_x, top_y = 0, mid_y, dy;
			int right_x, bottom_y = 0;
			double start_White, delta_White;
			double start_Chroma, delta_Chroma;
			Color pixel, start_pixel;

			gradient_flag = false;
			toolStripStatusLabel.Text = "";
			gradient_color = ColorPicker ("Select left-center fill", gradient_color);
			end_White = ReadInt ("Right edge White", 0, end_White, 254);
			end_Chroma = ReadInt ("Top edge Chroma", 0, end_Chroma, 254);
			HCW hcw = new HCW (gradient_color);

			// last clicked pixel location
			start_pixel = current_image.GetPixel (current_x, current_y);
			mid_y = current_y;
			for (right_x = current_x; right_x < (current_image.Width - 1); ++right_x)
			{
				Range (start_pixel, right_x, ref top_y, ref mid_y, ref bottom_y);
				if (start_pixel != current_image.GetPixel (right_x + 1, mid_y))
					break;
			}

			mid_y = current_y;
			for (left_x = current_x; left_x > 0; --left_x)
			{
				Range (start_pixel, left_x, ref top_y, ref mid_y, ref bottom_y);
				if (start_pixel != current_image.GetPixel (left_x - 1, mid_y))
					break;
			}

			start_White = hcw.White;
			delta_White = (end_White - start_White) / (right_x - left_x);
			if (2 < delta_White)
				delta_White = 2;
			if (-2 > delta_White)
				delta_White = -2;

			// Loop through the range pixels to create gradient.
			for (; left_x <= right_x; left_x++)
			{
				Range (start_pixel, left_x, ref top_y, ref mid_y, ref bottom_y);
				start_Chroma = gradient_color.Chroma;
				delta_Chroma = (end_Chroma - start_Chroma) / (top_y - mid_y);
				if (2 < delta_Chroma)
					delta_Chroma = 2;
				if (-2 > delta_Chroma)
					delta_Chroma = -2;
				hcw.Chroma = (byte) (start_Chroma);
				hcw.White = (byte) (start_White);
				pixel = hcw.RGB ();
				current_image.SetPixel (left_x, mid_y, pixel);
				for (dy = 1; (mid_y + dy) <= top_y; ++dy)
				{
					start_Chroma += delta_Chroma;
					hcw.Chroma = (byte) (start_Chroma);
					pixel = hcw.RGB ();
					current_image.SetPixel (left_x, mid_y + dy, pixel);
					if ((mid_y - dy) >= bottom_y)
						current_image.SetPixel (left_x, mid_y - dy, pixel);
				}
				start_White += delta_White;
			}
			UpdateImage ();
		}

		private void Range (Color start_pixel, int x, ref int top_y, ref int mid_y, ref int bottom_y)
		{
			// Calculate range in y for a given x of color start for gradient area search
			for (top_y = mid_y; top_y < (current_image.Height - 1); ++top_y)
			{
				if (start_pixel != current_image.GetPixel (x, top_y + 1))
					break;
			}
			for (bottom_y = mid_y; bottom_y > 0; --bottom_y)
			{
				if (start_pixel != current_image.GetPixel (x, bottom_y - 1))
					break;
			}
			mid_y = bottom_y + (top_y - bottom_y) / 2;
		}

		private void textMenuItem_Click (object sender, EventArgs e)
		{
			text_flag = true;
			toolStripStatusLabel.Text = "Click the initial starting point for inserting text.";
		}

		public void TextPrint ()
		{
			text_flag = false;
			toolStripStatusLabel.Text = "";
			RotateText text_form = new RotateText ();
			text_form.ShowDialog (this);
		}

		#endregion

		#region View Menu Items

		private void StatusBarMenuItem_Click (object sender, EventArgs e)
		{
			statusStrip.Visible = statusBarMenuItem.Checked;
		}

		#endregion

		#region Tools Menu Items

		private void colorPickerMenuItem_Click (object sender, EventArgs e)
		{
			current_color = ColorPicker ("Current", current_color);
		}

		public HCW ColorPicker (string str, HCW hcw)
		{
			Picker picker_form = new Picker ();
			picker_form.Value = hcw;
			picker_form.Text = "Color Picker - " + str;
			picker_form.ShowDialog (this);
			return picker_form.Value;
		}

		private void memoryUsedMenuItem_Click (object sender, EventArgs e)
		{
			long used;

			used = GC.GetTotalMemory (true);
			MessageBox.Show (String.Format ("{0:#,0} bytes\r\n{1:#,0} change", used, used - memory_last_used), "Memory used");
			memory_last_used = used;
		}

		private void analyzeSelectHueMenuItem_Click (object sender, EventArgs e)
		{
			// Generate analyze bitmaps around a specific current hue
			int[] hue = new int[360], chroma = new int[256], white = new int[256];
			int x = 0, y = 0;
			HCW hcw = new HCW (Color.Red);

			if (null == current_image.Bitmap)
			{
				MessageBox.Show ("No image!", "Analyze Select Hue");
				return;
			}

			Cursor.Current = Cursors.WaitCursor;

			// Loop through the images pixels
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));
					int select_hue = (hcw.Hue - current_color.Hue + 180 + 3600) % 3600;
					if (select_hue < 0 || select_hue > 359)
						continue;

					chroma[hcw.Chroma]++;
					white[hcw.White]++;
					// Skip near black and near white
					if (5 < hcw.Chroma && 250 > hcw.White)
						hue[select_hue]++;
				}
			}
			DisplayAnalyze ("Selected Hue Distribution", hue);
			DisplayAnalyze ("Selected Chroma Distribution", chroma);
			DisplayAnalyze ("Selected White Distribution", white);
		}

		private void analyzeHCWDistributionMenuItem_Click (object sender, EventArgs e)
		{
			// Generate analyze bitmaps
			int[] hue = new int[360], chroma = new int[256], white = new int[256];
			int x = 0, y = 0;
			HCW hcw = new HCW (Color.Red);

			if (null == current_image.Bitmap)
			{
				MessageBox.Show ("No image!", "Analyze HCW Distribution");
				return;
			}

			Cursor.Current = Cursors.WaitCursor;

			// Loop through the images pixels
			for (x = 0; x < current_image.Width; x++)
			{
				for (y = 0; y < current_image.Height; y++)
				{
					hcw.RGB (current_image.GetPixel (x, y));
					chroma[hcw.Chroma]++;
					white[hcw.White]++;
					// Skip near black and near white
					if (5 < hcw.Chroma && 250 > hcw.White)
						hue[hcw.Hue / 10]++;
				}
			}
			DisplayAnalyze ("Hue Distribution", hue);
			DisplayAnalyze ("Chroma Distribution", chroma);
			DisplayAnalyze ("White Distribution", white);
		}

		private void DisplayAnalyze (string text, int[] range)
		{
			// common function used to display analyze bitmaps
			Blank_Form analyze_form;
			Bitmap image = new Bitmap (range.GetLength (0) * 2, 256);
			int x, y, t;
			Color pixel;
			int bottom, top;
			HCW hcw = new HCW (Color.Red);

			analyze_form = new Blank_Form ();
			analyze_form.MdiParent = this;
			analyze_form.ClientSize = new Size (image.Width, image.Height);

			// Clipping chroma and white but not hue
			if (text.Contains ("Hue"))
			{
				bottom = 0;
				top = range.GetLength (0);
				hcw.RGB (Color.Red);
			}
			else
			{
				bottom = 2;
				top = range.GetLength (0) - 2;
			}

			// Normalize range
			y = 1;
			for (x = bottom; x < top; x++)
				if (y < range[x])
					y = range[x];
			for (x = 0; x < range.GetLength (0); x++)
			{
				t = range[x] * 255 / y;
				// make sure they see at least one if there are any
				if (0 == t && 0 < range[x])
					range[x] = 1;
				else
					range[x] = t;
			}

			for (x = 0; x < bottom; x++)
				if (255 < range[x])
					range[x] = 255;

			for (x = top; x < range.GetLength (0); x++)
				if (255 < range[x])
					range[x] = 255;

			// Loop through the pixels to make a black background and grid.
			for (x = 0; x < image.Width; x++)
			{
				for (y = 0; y < image.Height; y++)
				{
					if (0 == (x % 16) || 0 == (y % 16))
						if (0 == (x % 64) || 0 == (y % 64))
							pixel = Color.White;
						else
							pixel = Color.Gray;
					else
						pixel = Color.Black;
					image.SetPixel (x, 255 - y, pixel);
				}
			}

			// Loop through the pixels to create bitmap.
			for (x = 0; x < image.Width; x += 2)
			{
				if (text.Contains ("Hue"))
					if (text.Contains ("Select"))
						hcw.Hue = (x / 2) + current_color.Hue - 180;
					else
						hcw.Hue = (short) (x * 5);
				for (y = 0; y < range[x / 2]; y++)
				{
					image.SetPixel (x, 255 - y, hcw.RGB ());
					image.SetPixel (x + 1, 255 - y, hcw.RGB ());
				}
			}
			analyze_form.BackgroundImage = image;
			analyze_form.Text = text;
			analyze_form.Show ();
		}

		#endregion

		#region Window Menu Items

		private void CascadeMenuItem_Click (object sender, EventArgs e)
		{
			LayoutMdi (MdiLayout.Cascade);
		}

		private void TileVerticalMenuItem_Click (object sender, EventArgs e)
		{
			LayoutMdi (MdiLayout.TileVertical);
		}

		private void TileHorizontalMenuItem_Click (object sender, EventArgs e)
		{
			LayoutMdi (MdiLayout.TileHorizontal);
		}

		private void ArrangeIconsMenuItem_Click (object sender, EventArgs e)
		{
			LayoutMdi (MdiLayout.ArrangeIcons);
		}

		private void CloseAllMenuItem_Click (object sender, EventArgs e)
		{
			DiscardChanges ();
			CloseForms ();
		}

		#endregion

		#region Help Menu Items

		private void aboutMenuItem_Click (object sender, EventArgs e)
		{
			AboutBox frm = new AboutBox ();
			frm.ShowDialog ();
		}

		private void localContentsMenuItem_Click (object sender, EventArgs e)
		{
			string path = AppDomain.CurrentDomain.BaseDirectory + "\\HCW-Modify.html";
			if (!File.Exists (path))
			{
				string[] lines = { "Open online location", "and save a local copy to this", "location." };
				System.IO.File.WriteAllLines (path, lines);
			}
			Process process = new Process ();
			process.StartInfo.FileName = path;
			bool result = process.Start ();
		}

		private void contentsMenuItem_Click (object sender, EventArgs e)
		{
			Process process = new Process ();
			process.StartInfo.FileName = "http://www.frank-t-clark.com/Professional/Papers/ColorHCW/HCW-Modify.html";
			bool result = process.Start ();
		}

		#endregion

		#region Common Methods

		private void OpenFileEnable (bool flag)
		{
			editMenu.Enabled = flag;
			revertToOriginalMenuItem.Enabled = flag;
			closeMenuItem.Enabled = flag;
			saveAsMenuItem.Enabled = flag;
		}

		private void ToolFileEnable ()
		{
			toolStripStatusLabel.Text = current_file_name;
			editMenu.Enabled = true;
			closeMenuItem.Enabled = true;
			saveAsMenuItem.Enabled = true;
		}

		private void DiscardChanges ()
		{
			if (modified_flag)
			{
				if (DialogResult.Yes == MessageBox.Show (
				"Save modified image?",
				"Modified image not saved",
				MessageBoxButtons.YesNo,
				MessageBoxIcon.Question,
				MessageBoxDefaultButton.Button1))
					FileSaveAs ();
			}
			modified_flag = false;
		}

		private int ReadInt (string prompt, int minimum, int start, int maximum)
		{

			InputInt frm = new InputInt ();
			frm.Prompt = prompt;
			frm.Minimum = minimum;
			frm.Maximum = maximum;
			frm.Start = start;
			frm.ShowDialog ();
			start = frm.Start;
			return start;
		}

		#endregion

	}
}
