﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using HCW_Class;

namespace HCW_Modify.Forms
{
	public partial class Picker:Form
	{

		MDIMain mdiMain;

		// Properties
		HCW hcw_value = new HCW ();

		/// <summary>
		/// Value is the start and end hcw value
		/// </summary>
		public HCW Value
		{
			get
			{
				return hcw_value;
			}
			set
			{
				hcw_value = value;
			}
		}

		private bool updating = false;
		private bool lockChroma = false;
		private bool lockWhite = false;
		private HCW TSave=new HCW();	// save value for lock ratio

		public Picker ()
		{
			InitializeComponent ();
			this.Font = SystemFonts.MessageBoxFont;
		}

		private void Picker_Load (object sender, EventArgs e)
		{
			mdiMain = (MDIMain) this.Owner;
			if (mdiMain.transform_flag)
				if (mdiMain.transform_White)
				{
					C_numericUpDown.Enabled = false;
					C_trackBar.Enabled = false;
					C_button.Enabled = false;

				}
				else
				{
					W_numericUpDown.Enabled = false;
					W_trackBar.Enabled = false;
					W_button.Enabled = false;
				}
			UpdateForm ();
		}

		private void UpdateForm ()
		{
			if (updating)
				return;
			updating = true;

			// Special action when called by transform dialog
			if (mdiMain.transform_flag)
				if (mdiMain.transform_White)
				{
					hcw_value.C = 255;
				}
				else
				{
					hcw_value.W = 0;
				}

			HCW hcw_save = new HCW (hcw_value);
			HCW hcw = new HCW (hcw_save);
			Color rgb = hcw.RGB ();
			H_numericUpDown.Value = hcw.Hue;
			C_numericUpDown.Value = hcw.Chroma;
			W_numericUpDown.Value = hcw.White;
			H_trackBar.Value = hcw.Hue;
			C_trackBar.Value = hcw.Chroma;
			W_trackBar.Value = hcw.White;
			R_numericUpDown.Value = rgb.R;
			G_numericUpDown.Value = rgb.G;
			B_numericUpDown.Value = rgb.B;
			RGB_pictureBox.BackColor = rgb;

			// Create bitmaps
			int x, y;

			hcw.iHCW (hcw_save);
			Bitmap image = new Bitmap (10, 256);
			// Loop through the pixels to create Chroma bitmap.
			for (y = image.Height - 1; y >= 0; y--)
			{
				hcw.Chroma = (byte) y;
				rgb = hcw.RGB ();
				for (x = 0; x < image.Width; x++)
					image.SetPixel (x, image.Height - 1 - y, rgb);
			}
			C_pictureBox.BackgroundImage = image;

			hcw.iHCW (hcw_save);
			image = new Bitmap (10, 256);
			// Loop through the pixels to create White bitmap.
			for (y = 0; y < image.Height; y++)
			{
				hcw.White = (byte) y;
				rgb = hcw.RGB ();
				for (x = 0; x < image.Width; x++)
					image.SetPixel (x, image.Height - 1 - y, rgb);
			}
			W_pictureBox.BackgroundImage = image;

			hcw_value = hcw_save;

			updating = false;
		}

		private int NextStep (int value)
		{
			value++;
			int i = 31 + value - value % 32;
			if (255 < i)
				return 0;
			return i;
		}

		private void H_button_Click (object sender, EventArgs e)
		{
			hcw_value.Hue = (short) (hcw_value.Hue + 600 - hcw_value.Hue % 600);
			UpdateForm ();
		}

		private void C_button_Click (object sender, EventArgs e)
		{
			LockChroma (NextStep (hcw_value.Chroma));
			UpdateForm ();
		}

		private void W_button_Click (object sender, EventArgs e)
		{
			LockWhite (NextStep (hcw_value.White));
			UpdateForm ();
		}

		private void R_button_Click (object sender, EventArgs e)
		{
			Color rgb = hcw_value.RGB ();
			hcw_value.RGB (NextStep (rgb.R), rgb.G, rgb.B);
			UpdateForm ();
		}

		private void G_button_Click (object sender, EventArgs e)
		{
			Color rgb = hcw_value.RGB ();
			hcw_value.RGB (rgb.R, NextStep (rgb.G), rgb.B);
			UpdateForm ();
		}

		private void B_button_Click (object sender, EventArgs e)
		{
			Color rgb = hcw_value.RGB ();
			hcw_value.RGB (rgb.R, rgb.G, NextStep (rgb.B));
			UpdateForm ();
		}

		private void H_trackBar_Scroll (object sender, EventArgs e)
		{
			hcw_value.Hue = (short) H_trackBar.Value;
			UpdateForm ();
		}

		private void C_trackBar_Scroll (object sender, EventArgs e)
		{
			LockChroma (C_trackBar.Value);
			UpdateForm ();
		}

		private void W_trackBar_Scroll (object sender, EventArgs e)
		{
			LockWhite (W_trackBar.Value);
			UpdateForm ();
		}

		private void HCW_pictureBox_Click (object sender, EventArgs e)
		{
			int x, y;

			x = ((MouseEventArgs) e).X * 255 / ((PictureBox) sender).Size.Width;
			y = ((MouseEventArgs) e).Y * 255 / ((PictureBox) sender).Size.Height;
			Bitmap image = new Bitmap (HCW_pictureBox.BackgroundImage);
			hcw_value.RGB (image.GetPixel (x, y));
			UpdateForm ();
		}

		private void H_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			hcw_value.Hue = (short) H_numericUpDown.Value;
			UpdateForm ();
		}

		private void C_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			LockChroma ((int) C_numericUpDown.Value);
			UpdateForm ();
		}

		private void W_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			LockWhite ((int) W_numericUpDown.Value);
			UpdateForm ();
		}

		private void R_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			hcw_value.RGB ((byte) R_numericUpDown.Value, (byte) G_numericUpDown.Value, (byte) B_numericUpDown.Value);
			UpdateForm ();
		}

		private void G_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			hcw_value.RGB ((byte) R_numericUpDown.Value, (byte) G_numericUpDown.Value, (byte) B_numericUpDown.Value);
			UpdateForm ();
		}

		private void B_numericUpDown_ValueChanged (object sender, EventArgs e)
		{
			hcw_value.RGB ((byte) R_numericUpDown.Value, (byte) G_numericUpDown.Value, (byte) B_numericUpDown.Value);
			UpdateForm ();
		}

		private void OK_button_Click (object sender, EventArgs e)
		{
			Close ();
		}

		// lock chroma ratio checkbox
		private void checkBox1_CheckedChanged (object sender, EventArgs e)
		{
			if (checkBox1.Checked)
				TSave.iHCW (hcw_value);
			else
				lockChroma = false;
		}

		// lock white ratio checkbox
		private void checkBox2_CheckedChanged (object sender, EventArgs e)
		{
			if (checkBox2.Checked)
				TSave.iHCW (hcw_value);
			else
				lockWhite = false;
		}

		// transform chroma if ratio is locked
		private void LockChroma (int Chroma)
		{
			if (updating)
				return;
			if (checkBox1.Checked)
			{
				if (!lockChroma)
				{
					lockChroma = true;
					lockWhite = false;
					TSave.iHCW (hcw_value);
				}
				hcw_value.Chroma = TSave.Chroma;
				hcw_value.White = TSave.White;
				hcw_value.TChroma = Chroma;
			}
			else
			{
				hcw_value.Chroma = Chroma;
				TSave.iHCW (hcw_value);
			}
		}

		// transform white if ratio is locked
		private void LockWhite (int White)
		{
			if (updating)
				return;
			if (checkBox2.Checked)
			{
				if (!lockWhite)
				{
					lockWhite = true;
					lockChroma = false;
					TSave.iHCW (hcw_value);
				}
				hcw_value.Chroma = TSave.Chroma;
				hcw_value.White = TSave.White;
				hcw_value.TWhite = White;
			}
			else
			{
				hcw_value.White = White;
				TSave.iHCW (hcw_value);
			}
		}
	}
}
