﻿using System;
using System.Drawing;

namespace HCW_Class
{
	/// <summary> Alpha, Hue, Chroma, White class
	/// </summary>
	public class HCW
	{
		#region Data
		byte _alpha = 255;
		short _hue = 0;
		byte _chroma = 0;
		byte _white = 0;
		#endregion
		#region Properties A, Alpha, H, Hue, C, Chroma, W, and White
		/// <summary> A (Alpha) is a value restricted to the range 0 to 255
		/// where 0 = transparent and 255 equals opaque.
		/// </summary>
		public byte A
		{
			get
			{
				return _alpha;
			}
			set
			{
				_alpha = Restrict (value);
			}
		}
		/// <summary> Alpha is a value restricted to the range 0 to 255
		/// where 0 = transparent and 255 equals opaque.
		/// </summary>
		public int Alpha
		{
			get
			{
				return _alpha;
			}
			set
			{
				_alpha = Restrict (value);
			}
		}
		/// <summary> H (Hue) is a decidegree value in the range 0 to 3599
		/// where 0 = Red, 600 = Yellow,
		/// 1200 = Green, 1800 = Cyan,
		/// 2400 = Blue, 3000 = Magenta.
		/// </summary>
		public short H
		{
			get
			{
				return _hue;
			}
			set
			{
				_hue = HRestrict (value);
			}
		}
		/// <summary> Hue is a decidegree value in the range 0 to 3599
		/// where 0 = Red, 600 = Yellow,
		/// 1200 = Green, 1800 = Cyan,
		/// 2400 = Blue, 3000 = Magenta.
		/// </summary>
		public int Hue
		{
			get
			{
				return _hue;
			}
			set
			{
				_hue = HRestrict (value);
			}
		}
		/// <summary> C (Chroma) is restricted to the range 0 to 255 where
		/// 0 = Black, 255 = Maximum color
		/// The White value will be set less than or equal to the Chroma value.
		/// If White = Chroma the color is a level of grey.
		/// </summary>
		public byte C
		{
			get
			{
				return _chroma;
			}
			set
			{
				Chroma = Restrict (value);
				WRestrict ();
			}
		}
		/// <summary> Chroma is restricted to the range 0 to 255 where
		/// 0 = Black, 255 = Maximum color
		/// The White value will be set less than or equal to the Chroma value.
		/// If White = Chroma the color is a level of grey.
		/// </summary>
		public int Chroma
		{
			get
			{
				return _chroma;
			}
			set
			{
				_chroma = Restrict (value);
				WRestrict ();
			}
		}
		/// <summary> Chroma is restricted to the range 0 to 255 where
		/// 0 = Black, 255 = Maximum color
		/// The White value will transformed/scaled based on the change in Chroma.
		/// </summary>
		public int TChroma
		{
			get
			{
				return _chroma;
			}
			set
			{
				int former = _chroma;
				_chroma = Restrict (value);
				if (_chroma != former)
				{
					if (0 == former)
						_white = _chroma;
					else
						_white = (byte) (_white * _chroma / former);
				}
			}
		}
		/// <summary> W (White) is restricted to the range 0 to the Chroma value where
		/// 0 = a pure color.
		/// The White value will be set less than or equal to the Chroma value.
		/// If White = Chroma the color is a level of grey.
		/// </summary>
		public byte W
		{
			get
			{
				return _white;
			}
			set
			{
				_white = Restrict (value);
				WRestrict ();
			}
		}
		/// <summary> White value is restricted to the range 0 to the Chroma value where
		/// 0 = a pure color.
		/// The White value will be set less than or equal to the Chroma value.
		/// If White = Chroma the color is a level of grey.
		/// </summary>
		public int White
		{
			get
			{
				return _white;
			}
			set
			{
				_white = Restrict (value);
				WRestrict ();
			}
		}
		/// <summary> White value is restricted to the range 0 to the 255 value where
		/// 0 = a pure color.
		/// The Chroma value will transformed/scaled based on the change in White.
		/// </summary>
		public int TWhite
		{
			get
			{
				return _white;
			}
			set
			{
				int former = _white;
				_white = Restrict (value);
				if (_white != value)
				{
					if (255 == former)
						_chroma = _white;
					else
						_chroma = (byte) (255 - ((255 - _chroma) * (255 - _white) / (255 - former)));
				}
			}
		}
		#endregion
		#region Constructors
		/// <summary> Initialize to Color.Black
		/// </summary> 
		public HCW ()
		{
			Alpha = 255;
			Hue = 0;
			Chroma = 0;
			White = 0;
		}
		/// <summary> Initialize to new HCW values
		/// </summary>
		/// <param name="hcw">HCW value</param>
		public HCW (HCW hcw)
		{
			Alpha = hcw.Alpha;
			Hue = hcw.Hue;
			Chroma = hcw.Chroma;
			White = hcw.White;
		}
		/// <summary> Initialize to new HCW individual values
		/// </summary>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public HCW (int hue, int chroma, int white)
		{
			Alpha = 255;
			Hue = hue;
			Chroma = chroma;
			White = white;
		}
		/// <summary> Initialize to new AHCW individual values
		/// </summary>
		/// <param name="alpha">Alpha in the range 0 to 255</param>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public HCW (int alpha, int hue, int chroma, int white)
		{
			Alpha = alpha;
			Hue = hue;
			Chroma = chroma;
			White = white;
		}
		/// <summary>  Initialize to int value
		/// </summary> 
		/// <param name="ihcw">The int value to convert to the HCW value</param> 
		public HCW (int ihcw)
		{
			iHCW (ihcw);
		}
		/// <summary> Initialize to long value
		/// </summary> 
		/// <param name="ihcw">The long value to convert to the AHCW value</param> 
		public HCW (long ihcw)
		{
			iHCW (ihcw);
		}
		/// <summary> Initialize to Color value
		/// </summary> 
		/// <param name="rgb">The RGB value to convert to the HCW value</param> 
		public HCW (Color rgb)
		{
			RGB (rgb);
		}
		#endregion
		#region Private Helper Methods
		/// <summary> Restrict a value to the range 0 to 255
		/// </summary>
		static byte Restrict (int value)
		{
			return (byte) (value > 255 ? 255 : value < 0 ? 0 : value);
		}
		/// <summary> Restrict a value to the decidegree range 0 to 3599
		/// </summary>
		static short HRestrict (int value)
		{
			return (short) ((value + 3600) % 3600);
		}
		/// <summary> Restrict _white to be less than or equal to chroma
		/// </summary>
		 void WRestrict ()
		{
				_white = _white > _chroma ? _chroma : _white;
		}
		#endregion
		#region Public Methods
		/// <summary> set this HCW value from HCW structure
		/// </summary>
		/// <param name="hcw">HCW structure</param>
		public void iHCW (HCW hcw)
		{
			Alpha = hcw.Alpha;
			Hue = hcw.Hue;
			Chroma = hcw.Chroma;
			White = hcw.White;
		}
		/// <summary> set this HCW value from individual HCW values
		/// </summary>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public void iHCW (int hue, int chroma, int white)
		{
			Alpha = 255;
			Hue = hue;
			Chroma = chroma;
			White = white;
		}
		/// <summary> set this AHCW value from individual AHCW values
		/// </summary>
		/// <param name="alpha">Alpha in the range 0 to 255</param>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public void iHCW (int alpha, int hue, int chroma, int white)
		{
			Alpha = alpha;
			Hue = hue;
			Chroma = chroma;
			White = white;
		}
		/// <summary> return an integer equivalent value for this HCW instance
		/// </summary> 
		public int iHCW ()
		{
			return IHCW (_hue, _chroma, _white);
		}
		/// <summary> return a long equivalent value for this AHCW instance
		/// </summary> 
		public long lHCW ()
		{
			return IHCW (_alpha, _hue, _chroma, _white);
		}
		/// <summary> set this HCW value from an integer equivalent value for HCW
		/// </summary> 
		public void iHCW (int value)
		{
			Alpha = 255;
			Hue = (short) (value >> 16);
			Chroma = (byte) ((value >> 8) & 0xff);
			White = (byte) (value & 0xff);
		}
		/// <summary> set this AHCW value from a long equivalent value for AHCW
		/// </summary> 
		public void iHCW (long value)
		{
			Alpha = (byte) (value >> 32 & 0xff);
			Hue = (short) (value >> 16 & 0xffff);
			Chroma = (byte) ((value >> 8) & 0xff);
			White = (byte) (value & 0xff);
		}
		/// <summary> return an ARGB equivalent value converted from this AHCW instance
		/// </summary> 
		public Color RGB ()
		{
			return iRGB (_alpha, _hue, _chroma, _white);
		}
		/// <summary> set the AHCW equivalent value converted from ARGB Color structure
		/// </summary> 
		/// <param name="rgb">RGB structure</param>
		public void RGB (Color rgb)
		{
			RGB (rgb.A, rgb.R, rgb.G, rgb.B);
		}
		/// <summary> set the HCW equivalent value converted from individual RGB values
		/// </summary> 
		/// <param name="red">red value in the range 0 to 255</param>
		/// <param name="green">green value in the range 0 to 255</param>
		/// <param name="blue">blue value in the range 0 to 255</param>
		public void RGB (int red, int green, int blue)
		{
			RGB ((byte) red, (byte) green, (byte) blue);
		}
		/// <summary> set the AHCW equivalent value converted from individual ARGB values
		/// </summary> 
		/// <param name="alpha">alpha value in the range 0 to 255</param>
		/// <param name="red">red value in the range 0 to 255</param>
		/// <param name="green">green value in the range 0 to 255</param>
		/// <param name="blue">blue value in the range 0 to 255</param>
		public void RGB (int alpha, int red, int green, int blue)
		{
			RGB ((byte) alpha, (byte) red, (byte) green, (byte) blue);
		}
		/// <summary> set the HCW equivalent value converted from individual RGB values
		/// </summary> 
		/// <param name="red">red value in the range 0 to 255</param>
		/// <param name="green">green value in the range 0 to 255</param>
		/// <param name="blue">blue value in the range 0 to 255</param>
		public void RGB (byte red, byte green, byte blue)
		{
			int sextant, offset;

			if (red >= green)
				if (green >= blue)
					sextant = 0;
				else
					if (red >= blue)
						sextant = 5;
					else
						sextant = 4;
			else
				if (red >= blue)
					sextant = 1;
				else
					if (green >= blue)
						sextant = 2;
					else
						sextant = 3;

			switch (sextant)
			{
				case 0:	// 0 Red
					offset = red - blue;
					if (0 >= offset)
						_hue = 0;
					else
						_hue = (short) (((green - blue) * 600 + (offset - 1)) / offset);
					_chroma = red;
					_white = blue;
					break;
				case 1:	// 600 Yellow
					offset = green - blue;
					_hue = (short) (1200 - ((red - blue) * 600) / offset);
					_chroma = green;
					_white = blue;
					break;
				case 2:	// 1200 Green
					offset = green - red;
					_hue = (short) (1200 + (((blue - red) * 600 + (offset - 1)) / offset));
					_chroma = green;
					_white = red;
					break;
				case 3:	// 1800 Cyan
					offset = blue - red;
					_hue = (short) (2400 - ((green - red) * 600) / offset);
					_chroma = blue;
					_white = red;
					break;
				case 4:// 2400 Blue
					offset = blue - green;
					_hue = (short) (2400 + (((red - green) * 600 + (offset - 1)) / offset));
					_chroma = blue;
					_white = green;
					break;
				case 5:	// 3000 Magenta
				default:
					offset = red - green;
					_hue = (short) (3600 - ((blue - green) * 600) / offset);
					_chroma = red;
					_white = green;
					break;
			}
		}
		/// <summary> set the AHCW equivalent value converted from individual ARGB values
		/// </summary> 
		/// <param name="alpha">alpha value in the range 0 to 255</param>
		/// <param name="red">red value in the range 0 to 255</param>
		/// <param name="green">green value in the range 0 to 255</param>
		/// <param name="blue">blue value in the range 0 to 255</param>
		public void RGB (byte alpha, byte red, byte green, byte blue)
		{
			RGB (red, green, blue);
			_alpha = alpha;
		}
		/// <summary> return a string describing the current AHCW values
		/// </summary> 
		override public string ToString ()
		{
			return String.Format ("A: {0:D3} H: {1:D4} C: {2:D3} W: {3:D3}", Alpha, Hue, Chroma, White);
		}
		/// <summary> return a Bitmap (52x16) illustrating the current HCW values
		/// 3 16 x 16 solid bitmaps. 3*16+4 = 52
		/// Actual Hue, Max Chroma, Min White
		/// Actual Hue, Actual Chroma, Min White
		/// Actual Hue, Actual Chroma, Actual White
		/// </summary> 
		public Bitmap ToImage ()
		{
			Bitmap image;
			HCW hcw2 = new HCW ();
			int x, y;
			Color pixel;

			// 3 16 x 16 solid bitmaps. 3*16+4 = 52
			image = new Bitmap (52, 16);

			// Actual Hue, Max Chroma, Min White
			hcw2.iHCW (Hue, 255, 0);
			pixel = hcw2.RGB ();
			for (x = 0; x < 16; x++)
				for (y = 0; y < image.Height; y++)
					image.SetPixel (x, y, pixel);

			// Actual Hue, Actual Chroma, Min White
			hcw2.iHCW (Hue, Chroma, 0);
			pixel = hcw2.RGB ();
			for (x = 17; x < 32; x++)
				for (y = 0; y < image.Height; y++)
					image.SetPixel (x, y, pixel);

			// Actual Hue, Actual Chroma, Actual White
			hcw2.iHCW (Hue, Chroma, White);
			pixel = hcw2.RGB ();
			for (x = 33; x < 49; x++)
				for (y = 0; y < image.Height; y++)
					image.SetPixel (x, y, pixel);
			return image;
		}
		#endregion
		#region Public Static Methods
		/// <summary> static method to return an integer equivalent HCW value from individual HCW values
		/// </summary>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public static int IHCW (int hue, int chroma, int white)
		{
			return (int) (hue << 16) + (int) (chroma << 8) + (int) (white);
		}
		/// <summary> static method to return a long equivalent AHCW value from individual AHCW values
		/// </summary>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		/// <param name="alpha">Alpha in the range 0 to 255</param>
		public static long IHCW (int alpha, int hue, int chroma, int white)
		{
			return (long) (alpha << 32) + (long) (hue << 16) + (long) (chroma << 8) + (long) (white);
		}
		/// <summary> static method to return an RGB equivalent value
		/// </summary> 
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public static Color iRGB (int hue, int chroma, int white)
		{
			byte red, green, blue;
			int sextant, offset;

			sextant = hue / 600;	// integer sixth part of a circle
			offset = ((hue % 600) * (chroma - white) + 300) / 600;	// offset scaled to byte

			switch (sextant)
			{
				case 0:	// 0 Red
					red = (byte) chroma;
					if (chroma == white)
						green = (byte) white;
					else
						green = (byte) (offset + white);
					blue = (byte) white;
					break;
				case 1:	// 600 Yellow
					red = (byte) (chroma - offset);
					green = (byte) chroma;
					blue = (byte) white;
					break;
				case 2:	// 1200 Green
					red = (byte) white;
					green = (byte) chroma;
					blue = (byte) (offset + white);
					break;
				case 3:	// 1800 Cyan
					red = (byte) white;
					green = (byte) (chroma - offset);
					blue = (byte) chroma;
					break;
				case 4:// 2400 Blue
					red = (byte) (offset + white);
					green = (byte) white;
					blue = (byte) chroma;
					break;
				case 5:	// 3000 Magenta
				default:
					red = (byte) chroma;
					green = (byte) white;
					blue = (byte) (chroma - offset);
					break;
			}
			return Color.FromArgb (red, green, blue);
		}
		/// <summary> static method to return an ARGB equivalent value
		/// </summary> 
		/// <param name="alpha">Alpha in the range 0 to 255</param>
		/// <param name="hue">Hue in the range 0 to 3599</param>
		/// <param name="chroma">Chroma in the range 0 to 255</param>
		/// <param name="white">White in the range 0 to value of Chroma</param>
		public static Color iRGB (int alpha, int hue, int chroma, int white)
		{
			Color color = iRGB (hue, chroma, white);
			return Color.FromArgb (alpha, color.R, color.G, color.B);
		}
		#endregion
		/// <summary> Validation routine to verify conversion formulas
		/// </summary>
		public static void Validate ()
		{
			HCW hcw = new HCW ();
			int r, g, b;

			// round trip validate
			for (r = 0; r < 256; ++r)
				for (g = 0; g < 256; ++g)
					for (b = 0; b < 256; ++b)
					{
						hcw.RGB (r, g, b);
						Color pixel = hcw.RGB ();
						if (r != pixel.R)
							throw new Exception ("R Mismatch " + r + " != " + pixel.R);
						if (g != pixel.G)
							throw new Exception ("G Mismatch " + g + " != " + pixel.G);
						if (b != pixel.B)
							throw new Exception ("B Mismatch " + b + " != " + pixel.B);
					}
			hcw.RGB (Color.Red);
			// Validate all 1530 (255*6) unique primary colors in the 360 color wheel
			for (int x = 0; x < 1530; x++)
			{
				hcw.Hue = (3600 * x) / 1530;
				Color pixel = hcw.RGB ();
				switch (x / 255)
				{
					case 0:
						if (255 != pixel.R)
							throw new Exception ("R Mismatch 255 != " + pixel.R);
						if (x % 255 != pixel.G)
							throw new Exception ("G Mismatch " + x % 255 + " != " + pixel.G);
						if (0 != pixel.B)
							throw new Exception ("B Mismatch 0 != " + pixel.B);
						break;
					case 1:
						if (255 - x % 255 != pixel.R)
							throw new Exception ("R Mismatch " + (255 - x % 255) + " != " + pixel.R);
						if (255 != pixel.G)
							throw new Exception ("G Mismatch 255 != " + pixel.G);
						if (0 != pixel.B)
							throw new Exception ("B Mismatch 0 != " + pixel.B);
						break;
					case 2:
						if (0 != pixel.R)
							throw new Exception ("R Mismatch 0 != " + pixel.R);
						if (255 != pixel.G)
							throw new Exception ("R Mismatch 255 != " + pixel.R);
						if (x % 255 != pixel.B)
							throw new Exception ("B Mismatch " + x % 255 + " != " + pixel.B);
						break;
					case 3:
						if (0 != pixel.R)
							throw new Exception ("R Mismatch 0 != " + pixel.R);
						if (255 - x % 255 != pixel.G)
							throw new Exception ("G Mismatch " + (255 - x % 255) + " != " + pixel.G);
						if (255 != pixel.B)
							throw new Exception ("B Mismatch 255 != " + pixel.B);
						break;
					case 4:
						if (x % 255 != pixel.R)
							throw new Exception ("R Mismatch " + x % 255 + " != " + pixel.R);
						if (0 != pixel.G)
							throw new Exception ("G Mismatch 0 != " + pixel.G);
						if (255 != pixel.B)
							throw new Exception ("B Mismatch 255 != " + pixel.B);
						break;
					case 5:
						if (255 != pixel.R)
							throw new Exception ("R Mismatch 255 != " + pixel.R);
						if (0 != pixel.G)
							throw new Exception ("G Mismatch 0 != " + pixel.G);
						if (255 - x % 255 != pixel.B)
							throw new Exception ("B Mismatch " + (255 - x % 255) + " != " + pixel.B);
						break;
				}
			}
		}
	}
}
